/*
MyDAC III Project

Arduino DAC manager for a  modified "Weiliang AK4495 AK4118 V1.1" DAC board.

The Arduino UNO and some aditionnal hardware replace the board with the 8051 MCU controller and display.

IDE used is codebender.cc

-The DAC manager software provide menu screens on 16*2 LCD, IR remote, additional coax input (relay board with 4 inputs and 1 output) and soft control
over AK4118 and AK4495 features : filter selection, sound quality selection, mute, volume control, input selection
and some error handling. DSD is not implemented as it imply too much mods on the pcb. AK4118 is SPI driven
and AK4495 is I2C, this setup minimize the pcb changes around the 4495. The chips are version : AK4118A and AK4495EQ.

-Software is event driven and screen based. Main loop consists of scrutating for events and then
to execute the corresponding action according to the screen displayed : one same event may trig different actions
regarding the screen in which the event occured.
-Event to Screen to Action mapping is done through lookup tables (arrays) which delivers easily customizable solution.
-Remote control key to Event mapping is also done through lookup tables (arrays).
-Software provide a 'Trim volume' facility per input wich allow to balance the input level of each input
independantly of the DAC master volume; useful when sources have very different level.
-Some DAC parameters are stored in EEPROM in order to keep them during power off.
-Software use three interval counters, software implemented around variables Delay1, 2 and 3
-The two available interrupts of the UNO are used, one for rotary switch, other for AK4118 INT0.
-Overall solution works fine but :
	initialize sequence of 4118 and 4495 works, but may be rationnalized
	Sound Quality control of AK4495 (Reg8) is implemented, but difference in SQ hard to notice and AKM documentation tells very few about it,
	except that Reg8 shouldn'nt be written...
	
AIM65, aim65@free.fr
2015 - 2016
Final Version

*/

// Write here the software version to be displayed at power up
char* MyDACIII_Version[] = {"v12"};

// Use IRremote for the remote control
#include <IRremote.h>
#include <IRremoteInt.h>
// Use I2C and LCD_I2C libs to communicate with AK4495 and the LCD
// See comment about LiquiCrystal_I2C (Use YWRObot V1.1, not the default Codebender one)
//Use A5 & A4 on PortC for SDA & SCL I2C lines
#include <Wire.h>
#include <LiquidCrystal_I2C.h>
// Use SPI lib for 4 wires communications with AK4118
#include <SPI.h>
// Use EEPROM lib to keep some setup during power down
#include <EEPROM.h>


//**** Arduino Pin definition ****
//Rotary encoder with push button
#define Enc_SWI 4
#define Enc_CLK 2		//connected to UNO INT0
#define Enc_DAT 8

//Infrared Receiver
#define IR_Rcv 5

//Debug probe
#define Probe 9			//used fo debugging

//Reset signal of AK4118
#define PDN	7

//Interrupt of AK4118
#define AK4118_INT0	3	//connected to UNO INT1

//Input Selector board with 4 relays
#define Relay4 14		//PortC A0
#define Relay3 15		//PortC A1
#define Relay2 16		//PortC A2
#define Relay1 17		//PortC A3

//4 wires interface with AK4118
//CSN is chip select pin of 4118, pin 11 is MOSI, pin 12 is MISO,
//pin 13 is CLOCK. Pin 11,12 and 13 are defined by the Wire library
//Even if CSN is part of the 4 wires interface, it is not managed by the Wire library
//but by the software.
#define CSN 10

//***** IR remote control code definition (NAD SR5 remote control) ****
#define Remote_Mfg	NEC			//NEC is defined by IR lib
#define Repeat_IR	0xFFFFFFFF	//Repeat code
#define MuteK		0xE13EE31C	//Key 0
#define InK1		0xE13E51AE	//Key 1
#define InK2		0xE13E718E	//Key 2
#define InK3		0xE13E49B6	//Key 3
#define InK4		0xE13E6996	//Key 4
#define InK5		0xE13ED12E	//Key 5
#define InK6		0xE13EF10E	//Key 6
#define VolUp		0xE13EB946	//Key Scan forward
#define VolDown		0xE13E7986	//Key Scan reverse
#define Left		0xE13ECB34	//Key Left
#define Right		0xE13E2BD4	//Key Right
#define Up			0xE13E4BB4	//Key Up
#define Down		0xE13E8B74	//Key Down
#define EnterK		0xE13EA35C	//Key Enter
#define IR_Code_7	0xE13EC936	//Not used
#define IR_Code_8	0xE13EE916	//Not used
#define IR_Code_9	0xE13E19E6	//Not used
#define IR_Code_On	0xE13EA45B	//Not used
#define IR_Code_Off	0xE13E13EC	//Not used

//***** Event Identification ****
#define No_Event			0			//No event
#define Ev_De1				1			//Menu screen time out delay
#define Ev_De2				2			//Trim screen time out delay
#define Ev_De3				3			//Screen animation delay
#define Ev_BPS				0x10		//Button pressed during short time
#define Ev_BPL				0x11		//Button pressed during longer time		
#define Ev_BRR				0x12		//Button rotated right
#define Ev_BRL				0x13		//Button rotated left
#define Ev_VolPlus			0x20		//Increase Volume
#define Ev_VolMinus			0x21		//Decrease Volume
#define Ev_SqPlus			0x22		//Next Sound Quality requested
#define Ev_SqMinus			0x23		//Previous Sound Quality requested
#define Ev_FiPlus			0x24		//Next Filter requested
#define Ev_FiMinus			0x25		//Previous Filter requested
#define Ev_Mute				0x26		//Mute toggle
#define Ev_In0				0x30		//Input0 requested
#define Ev_In1				0x31		//Input1 requested
#define Ev_In2				0x32		//Input2 requested
#define Ev_In3				0x33		//Input3 requested
#define Ev_In4				0x34		//Input4 requested
#define Ev_In5				0x35		//Input5 requested
#define	Ev_Enter			0x36		//Enter Action is requested
#define Ev_4118_Error		0x40		//AK4118 generated event in case of error
#define Ev_4118_End_Error	0x41		//Software generated event when AK4118 error condition is cleared
#define Ev_4118_SfChange	0x42		//AK4118 generated event is case sampling frequency change detection


//***** Action Identification ****
#define Ac_Noac			0			//No action
#define	Ac_MS_anim		1			//Main Screen animation sequencer
#define	Ac_Mute_Toggle	2			//Toggle mute switch
#define	Ac_MS			0x10		//Go to Main Screen		
#define	Ac_VS			0x11		//Go to Volume Screen	
#define	Ac_SS			0x12		//Not used anymore	
#define	Ac_FS			0x13		//Go to Filter Screen	
#define	Ac_QS			0x14		//Go to Sound Quality Screen	
#define	Ac_TS			0x15		//Go to Trim volume Screen	
#define	Ac_TSnext		0x16		//Go to Next Trim volume Screen	
#define	AC_ES			0x17		//Not used yet
#define	Ac_VSplus		0x20		//Raise Volume
#define	Ac_VSminus		0x21		//Lower Volume
#define	Ac_QSplus		0x22		//Next Sound Quality
#define	Ac_QSminus		0x23		//Previous sound Quality
#define	Ac_FSplus		0x24		//Next Filter
#define	Ac_FSminus		0x25		//Previous Filter
#define	Ac_TSplus		0x26		//Raise Trim Volume
#define	Ac_TSminus		0x27		//Lower Trim Volume
#define	Ac_SSplus		0x28		//Next source
#define	Ac_SSminus		0x29		//Previous Source
#define	Ac_In0			0x30		//Switch to input 0
#define	Ac_In1			0x31		//Switch to input 1
#define	Ac_In2			0x32		//Switch to input 2
#define	Ac_In3			0x33		//Switch to input 3
#define	Ac_In4			0x34		//Switch to input 4
#define	Ac_In5			0x35		//Switch to input 5
#define Ac_4118_Err		0x40		//Handle AK4118 error
#define Ac_4118_End_Err	0x41		//Terminate the Error
#define	Ac_SfChange		0x42		//Sampling frequency change detected
#define Ac_Debug		0x50		//Debug

//***** Screen Identification ****
#define	MS			0	//Main Screen
#define VS			1	//Volume Screen
#define	SS			2	//Not used anymore
#define	QS			3	//Sound Quality Screen
#define FS			4	//Filter Screen
#define TS			5	//Volume Trim Screen


//**** Initial values for AKM chips registers

#define	AK4118_Reg0_I	0x23	//No BCU, MCK=PLL, Clock Mode 2 (PLL then Xtal)
#define AK4118_Reg1_I	0x58	//Stereo, I2s 24b, auto de-emphasis
#define	AK4118_Reg2_I	0
#define AK4118_Reg3_I	0
#define AK4118_Reg4_I	0xA5	//INT0 on AUTO, ULOCK, DTS, NoAudio
#define AK4118_Reg5_I	0xFF	//No INT1
#define AK4118_Reg20_I	0x47	//No GPIO, No FAST PLL
#define AK4118_Reg24_I	0x3A	//XMCK=0, RXDetect do not trig INT0

#define	AK4495_Reg0_I	0x87	//24b, I2S, no external filter, auto clock setting
#define AK4495_Reg1_I	0x02	//no de-emphasis (done by AK4118)
#define	AK4495_Reg2_I	0
#define AK4495_Reg5_I	0
#define AK4495_Reg6_I	0
#define AK4495_Reg7_I	0

#define AK4495_Addr		B00010010	//I2C adress set by internal hardware (bit6-bit2 : 00100), then CAD1, CAD0
//CAD1 is wired to H and CAD0 is wired to L. Thus I2C@ = 0x12

//**** EEPROM adresses of variables to store when DAC powered off
#define NV_TVOL0		0		//Trim Volume of Input 0
#define NV_TVOL1		1		//Trim Volume of Input 1
#define NV_TVOL2		2		//Trim Volume of Input 2
#define NV_TVOL3		3		//Trim Volume of Input 3
#define NV_TVOL4		4		//Trim Volume of Input 4
#define NV_TVOL5		5		//Trim Volume of Input 5
#define NV_Filter		6		//Filter choosed
#define NV_SoundQ		7		//Sound quality choosed

//**** AK4118 Error code definition ****
#define NoErr			0
#define RxErr			1
#define UlckErr			2
#define DtsErr			3
#define NoPcmErr		4
#define DatErr			5
#define NoAudioErr		6
#define SfErr			7

//**** Interupt handling states definition ****
#define Int_No			0
#define Int_Trig		1
#define Int_Ongoing		2
#define Int_End			3

//**** Other definition ****
#define None			0
#define Delay1_Val		2000		//Menu screen time out delay in ms
#define Delay2_Val		5000		//Trim screen time out delay in ms
#define Delay3_Val		250			//Screen animation delay in ms
#define Anim_Lenght 	5			//# of steps of the Play animation on Main screen
#define Max_Trim		12			//Trim volume max attenuation is -12dB
#define LCD_BackL_Min	80			//Value for minimum brightness
#define LCD_BackL_Max	160			//Value for maximum brightness
#define Nofreq			6			//Index value for non supported Sf


//**** Lookup tables for Events to Actions mapping per screen ****
//This allow to parametrize, for each screen, the events enabled and the
//corresponding actions to execute

// Main Screen
int MS_Event[] = {Ev_De3, Ev_Enter, Ev_BPS, Ev_BPL, Ev_BRR, Ev_BRL, Ev_VolPlus, Ev_VolMinus, Ev_SqPlus, Ev_SqMinus, Ev_FiPlus, Ev_FiMinus, Ev_Mute, Ev_In0, Ev_In1, Ev_In2, Ev_In3, Ev_In4, Ev_In5, Ev_4118_Error, Ev_4118_End_Error, Ev_4118_SfChange , 0xFF};
int MS_Action[] = {Ac_MS_anim, Ac_Debug, Ac_VS, Ac_TS, Ac_SSplus, Ac_SSminus, Ac_VSplus, Ac_VSminus, Ac_QS, Ac_QS, Ac_FS, Ac_FS, Ac_Mute_Toggle, Ac_In0, Ac_In1, Ac_In2, Ac_In3, Ac_In4, Ac_In5, Ac_4118_Err, Ac_4118_End_Err, Ac_SfChange};

//Volume Screen
int VS_Event[] = {Ev_De1 , Ev_Enter, Ev_BPS, Ev_BRR, Ev_BRL, Ev_VolPlus, Ev_VolMinus, Ev_Mute, Ev_4118_Error, Ev_4118_End_Error, Ev_4118_SfChange, 0xFF};
int VS_Action[] = {Ac_MS , Ac_MS, Ac_FS, Ac_VSplus, Ac_VSminus, Ac_VSplus, Ac_VSminus, Ac_Mute_Toggle, Ac_4118_Err, Ac_4118_End_Err, Ac_SfChange};

//Filter Screen
int FS_Event[] = {Ev_De1 , Ev_Enter, Ev_BPS, Ev_BRR, Ev_BRL, Ev_SqPlus, Ev_SqMinus, Ev_FiPlus, Ev_FiMinus, Ev_Mute, Ev_4118_Error, Ev_4118_End_Error, Ev_4118_SfChange, 0xFF};
int FS_Action[] = {Ac_MS , Ac_MS, Ac_QS, Ac_FSplus, Ac_FSminus, Ac_QS, Ac_QS, Ac_FSplus, Ac_FSminus, Ac_Mute_Toggle, Ac_4118_Err, Ac_4118_End_Err, Ac_SfChange};

//Sound Quality Screen
int QS_Event[] = {Ev_De1 , Ev_Enter, Ev_BPS, Ev_BRR, Ev_BRL, Ev_SqPlus, Ev_SqMinus, Ev_FiPlus, Ev_FiMinus, Ev_Mute, Ev_4118_Error, Ev_4118_End_Error, Ev_4118_SfChange, 0xFF};
int QS_Action[] = {Ac_MS , Ac_MS, Ac_MS, Ac_QSplus, Ac_QSminus, Ac_QSplus, Ac_QSminus, Ac_FS, Ac_FS, Ac_Mute_Toggle, Ac_4118_Err, Ac_4118_End_Err, Ac_SfChange};

//Trim Volume Screens (1 to 6)
int TS_Event[] = {Ev_De2 , Ev_Enter, Ev_BPS, Ev_BPL, Ev_BRR, Ev_BRL, Ev_Mute, Ev_4118_Error, Ev_4118_End_Error, Ev_4118_SfChange, 0xFF};
int TS_Action[] = {Ac_MS , Ac_MS, Ac_TSnext, Ac_MS, Ac_TSminus, Ac_TSplus, Ac_Mute_Toggle, Ac_4118_Err, Ac_4118_End_Err, Ac_SfChange};


//**** Lookup tables for sampling frequency detection and management ****
//Allow to identify the frequency in 4118 Reg#7 then identify value to display and the
//master clock (MCK) to generate in 4118 Reg#0 and also to lookup the trim volume to apply
//to each input in order to equalize the sound level between inputs. Not all supported freq used.
//

byte	AK4118_Sf[6] =	{0, 0x20, 0x80, 0xA0, 0xC0, 0xE0};
char*	Sf_Display[7] =	{" 44k", " 48k", " 88k", " 96k", "176k", "192k", "---k"};	//Unsupported frequency displayed as ---k
byte	AK4118_MCK[6] = {0x08, 0x08, 0, 0, 0x0C, 0x0C};								//This allow to set the MCK linked to the detected Sf range

//**** Input Identification for display ****
char* Input_Name [6] = {"USB  ", "TV   ", "ROKU ", "CD   ", "AIR  ", "COAX4"};		//USB connected to AK4118 RX#2, TV is optical on RX#0, 4 others are coax on RX#1

//**** Filter Identification for display ****
char* Filter_Name [10] = {"Sharp", "           ", "Sharp", "Short Delay", "Slow ", "           ", "Slow ", "Short Delay", "     ", "Super Slow "};

//**** Error messages for display ****
char* Error_msg[8] = {"Unknown   ", "No Input  ", "PLL UnLock", "DTS Stream", "No PCM    ", "DAT Stream", "No Audio  ", "New fs    "};	//10 char per error


//**** Various global variable definition ****
byte dummy;
int AK4118_Err_Code;			//error code
int Prev_AK4118_Error;			//previous error code
boolean	Pre_emph_status;		//pre emphasis status based on AK4118 detection
int	Trim_Volume[6];				//array to store the trim volume of the six inputs
int DAC_Vol;			 		//global displayed output level of DAC, from 0 (0dB) to -127 (-127dB)
boolean	Mute_SW;				//status of mute toggle switch managed by the remote control

byte AK4495_Reg1 = AK4495_Reg1_I;	//Value of AK4495 Reg1 has to be managed in RAM because it is modified by more than
//one function and the I2C read register capability is not implemented in this software

boolean refresh_MS;			//true to force refresh of the entire MS screen
boolean refresh_VS;			//true to force refresh of the data, not the mask, on the VS screen
boolean refresh_TS;			//true to force refresh of the data, not the mask, on the TS screen
boolean refresh_QS;			//true to force refresh of the data, not the mask, on the QS screen
boolean refresh_FS;			//true to force refresh of the data, not the mask, on the FS screen

boolean Save_NV_QS;			//true to request save new Sound Quality in EEPROM
boolean Save_NV_FS;			//true to request save new Filter in EEPROM
boolean Save_NV_TS;			//true to request save new Trim Volume in EEPROM

int Current_Screen;			//0 to 5
int Previous_Screen;		//0 to 5
int Current_Input;			//0 to 5
int Current_Freq;			//0 to 5
int Current_Filter;			//0 to 4
int Current_SoundQ;			//0 to 4
int Current_TScreen;		//0 to 5

volatile int	Int_Enc_Event;		//used to pass result from switch Interrupt Service Routine
volatile int	Int_AK4118;			//state # of AK4118 interrupt handling process
volatile byte	AK4118_Reg6_img;	//image of AK4118 Reg6
volatile byte	AK4118_Reg26_img;	//image of AK4118 Reg26
byte			AK4118_Reg7_img;	//image of AK4118 Reg7

byte Animation_ctr = 0;		//Animation counter for Main Screen display
unsigned long Delay1_Ctr;	//used to manage Delay1
unsigned long Delay2_Ctr;	//...
unsigned long Delay3_Ctr;	//...

//Declare lcd object (@=0x3F, 16 chars, 2 lines)
LiquidCrystal_I2C lcd(0x3F, 16, 2);

//Declare irrecv object
IRrecv irrecv(IR_Rcv);
decode_results results;		//Used by IR lib


//**** Initialization, executed once at startup ****
//***************************************************

void setup()
{
//**** initialize digital pin mode ****
	pinMode(Relay1, OUTPUT);
	pinMode(Relay2, OUTPUT);
	pinMode(Relay3, OUTPUT);
	pinMode(Relay4, OUTPUT);
	pinMode(CSN, OUTPUT);					//Chip Select for AK4118
	pinMode(PDN, OUTPUT);					//Power DowN for AK4118
	pinMode(Probe, OUTPUT);
	pinMode(AK4118_INT0, INPUT_PULLUP);

	pinMode(Enc_CLK, INPUT_PULLUP);
	pinMode(Enc_DAT, INPUT_PULLUP);
	pinMode(Enc_SWI, INPUT_PULLUP);

	digitalWrite(CSN, HIGH);
	digitalWrite(PDN, LOW);					//Reset AK4118

//**** StartUp Display ****
	lcd.noBacklight();
	lcd.init();
	lcd.clear();
	lcd.noCursor();

//**** Launch IR Lib ****
	irrecv.blink13(false);					//IR lib should'nt use pin13 as it is used by Wire lib
	irrecv.enableIRIn();
	digitalWrite(IR_Rcv, HIGH);				//Should activate PullUp...
	irrecv.resume();

//**** initialize SPI communication ****
	SPI.setClockDivider(SPI_CLOCK_DIV4);	//SPI clock is 4MHz (16/4), 4118's max is 5
	SPI.setDataMode (SPI_MODE3); 			//4118 Clock Phase is High, Clock Polarity is High (rising edge)
	SPI.setBitOrder (MSBFIRST);				//MSB sent first
	SPI.begin();							//Start SPI

//**** initialize EEPROM on first time MCU usage ****
	for (int i = 0; i < 5; i++)
	{
		if (EEPROM.read(NV_TVOL0 + i) > Max_Trim)
		{
			EEPROM.write(NV_TVOL0 + i, 0);
		}
	}
	if (EEPROM.read(NV_Filter) > 4)
	{
		EEPROM.write(NV_Filter, 0);
	}
	if (EEPROM.read(NV_SoundQ) > 4)
	{
		EEPROM.write(NV_SoundQ, 0);
	}


//**** DAC startup parameters ****
	Current_Screen = MS;						//Main screen fired after startup one
	Previous_Screen = TS;						//force MS to be refreshed
	Current_Input = 1;							//startup input; 1 is Optical = TV
	DAC_Vol = -4;								//startup volume is -4dB
	Current_Filter = EEPROM.read(NV_Filter);	//load settings stored in EEPROM
	Current_SoundQ = EEPROM.read(NV_SoundQ);
	for (int i = 0; i < 6; i++)
	{
		Trim_Volume[i] = EEPROM.read(NV_TVOL0 + i);
	}

//**** Launch sequence ****
	lcd.print("    MyDAC III");
	lcd.setCursor(0, 1);
	lcd.print(MyDACIII_Version[0]);
	for (int i = 0; i < 16; i++)
	{
		lcd.scrollDisplayRight();
	}
	lcd.backlight();
	delay(100);
	for(int i = 0; i < 16; i++)
	{
		lcd.scrollDisplayLeft();
		delay(100);
	}

	digitalWrite(PDN, HIGH)	;					//release reset pin of AK4118
	delay(50);									//wait 4118 to start (if necessary...?)
	AK4118_Write(0, 0);							//RSTN & PWN
	delay(50);
	AK4118_Write(0, 1);							//release RSTN
	delay(50);
	AK4118_Write(0, AK4118_Reg0_I);				//release PWN and setup Reg0
	AK4118_Write(1, AK4118_Reg1_I);
	AK4118_Write(2, AK4118_Reg2_I);
	AK4118_Write(3, AK4118_Reg3_I);
	AK4118_Write(4, AK4118_Reg4_I);
	AK4118_Write(5, AK4118_Reg5_I);
	AK4118_Write(0x20, AK4118_Reg20_I);
	AK4118_Write(0x24, AK4118_Reg24_I);

	Set_Dac_In();								//set the input
	delay(200);

	AK4495_Write(0, AK4495_Reg0_I);				//release RSTN
	AK4495_Write(0, AK4495_Reg0_I);				//setup Reg0 once RSTN released
	AK4495_Write(1, AK4495_Reg1);
	AK4495_Write(2, AK4495_Reg2_I);
	AK4495_Write(5, AK4495_Reg5_I);
	AK4495_Write(6, AK4495_Reg6_I);
	AK4495_Write(7, AK4495_Reg7_I);
	Set_Dac_Filter();							//Set the filter
	Set_Dac_SoundQ();							//set the sound quality
	Set_Dac_Vol();								//set the volume
	Sf_Change();								//update Sf
	dummy = AK4118_Read(6);
	dummy = AK4118_Read(0x26);

//**** initialize interrupt handling
//Old method, compatible with https://codebender.cc which is the IDE for this project.
//Use of digitalPinToInterrupt(pin) is not supported
//thus int# should be used rather than pin#: 0 for INT0 on pin2 and 1 for INT1 on pin3
	Int_AK4118 = Int_No;
	Int_Enc_Event = No_Event;
	attachInterrupt(1, AK4118_Int, RISING);
	attachInterrupt(0, Switch_Int, CHANGE);
	delay(2000);

	if (digitalRead(AK4118_INT0))			//if error at startup
	{
		AK4118_Int();						//handle error with the interrupt handler
	}
}

//**** Switch Interrupt Service Routine ****
//Get direction of encoder rotation based on Clock and Data lines
//*******************************************
void Switch_Int(void)
{
	delayMicroseconds(7000);					//wait for debounce
	{
		if(digitalRead(Enc_CLK) == digitalRead(Enc_DAT))
		{
			Int_Enc_Event = Ev_BRL;
		}
		else
		{
			Int_Enc_Event = Ev_BRR;
		}
	}
}

//**** AK4118 Interrupt Service Routine ****
//Take snapshot of 4118 registers and initiate eror handling
//*******************************************
void AK4118_Int(void)
{
	if (Int_AK4118 == Int_No)							//Do not trig if error already ongoing
	{
		Int_AK4118 = Int_Trig;							//Set status at Interrupt triggered
		AK4118_Reg26_img = AK4118_Read(0x26);			//Reg26 should be read soon after INT0 as STC bit stay high during one subframe only
		AK4118_Reg6_img = AK4118_Read(6);
	}
}


//**** Infrared Event detection function ****
//Reply with an Event Code corresponding to the key pressed
//Event Code is No_Event when no key pressed
//*******************************************
int IR_Event_Detect(void)
{
	static unsigned long IR_Code_Enabled[] = {MuteK, EnterK, InK1, InK2, InK3, InK4, InK5, InK6, VolUp, VolDown, Right, Left, Up, Down, 0xFF};
	static int Event_Mapped[] = {Ev_Mute, Ev_Enter, Ev_In0, Ev_In1, Ev_In2, Ev_In3, Ev_In4, Ev_In5, Ev_VolPlus, Ev_VolMinus, Ev_SqPlus, Ev_SqMinus, Ev_FiPlus, Ev_FiMinus};
	static unsigned long	Previous_IR_Code;

	unsigned long IR_Code = None;
	int Event = No_Event;

	if (irrecv.decode(&results))
	{
		if (results.decode_type == Remote_Mfg)	//Pre filter IR Code on manufacturer
		{
			IR_Code = results.value;			//Get Code

			if (IR_Code == Repeat_IR & (Previous_IR_Code == VolUp | Previous_IR_Code == VolDown))			//enable repeat for volume keys
			{
				IR_Code = Previous_IR_Code;
			}
			int i = 0;
			do
			{
				if (IR_Code == IR_Code_Enabled[i])
				{
					Event = Event_Mapped[i];
				}
				i++;
			}
			while (Event == No_Event && IR_Code_Enabled[i] != 0xFF);
			Previous_IR_Code = IR_Code;
		}
		irrecv.resume(); 						//Get ready for next code
	}
	return (Event);
}

//**** Rotary Encoder Event detection function ****
//Reply with an Event Code corresponding to the usage on the encoder
//Event Code is No_Event when encoder not used
//Encoder has three wires : Clk, Data and Switch
//Function operate in interrupt mode and polling mode
//*************************************************
int Button_Event_Detect(void)
{
	int Event = No_Event;
	int Enc_switch = digitalRead(Enc_SWI);

	if(Int_Enc_Event)					//true (not zero) when interupt has occured
	{
		Event = Int_Enc_Event;
		Int_Enc_Event = No_Event;		//clear event detection
	}

	int PushDuration = 0;
	while (!Enc_switch)					//loop until button is released
	{
		delay(10);
		PushDuration++;					//measure duration
		Enc_switch = digitalRead(Enc_SWI);
	}
	if (PushDuration > 100)				//1 sec
	{
		Event = Ev_BPL;					//event is Button Push Long
	}
	else if (PushDuration > 11)			//110ms
	{
		Event = Ev_BPS;					//event is Button Push Short
	}
	return(Event);
}

//**** AK4118 error detection ****
//Detect one of the AK4118 supported error
//***************************************
byte Identify_AK4118_Error(void)
{
	byte Error = NoErr;

	if (!Current_Input_Status())				//current input disconnected
	{
		Error = RxErr;
	}
	else if (AK4118_Reg26_img & B00000010)		//STC bit on when SF or de-emphasize change
	{
		Error = SfErr;
	}
	else if (AK4118_Reg26_img & B00000001)		//DAT bit
	{
		Error = DatErr;
	}
	else if(AK4118_Reg6_img & B00001000)		//DTS bit
	{
		Error = DtsErr;
	}
	else if(AK4118_Reg6_img & B01000000)		//AUTO bit : non PCM auto detect
	{
		Error = NoPcmErr;
	}
	else if(AK4118_Reg6_img & B00000010)		//AUDION bit : non audio detect
	{
		Error = NoAudioErr;
	}
	else if(AK4118_Reg6_img & B00010000)		//UnLock bit (most frequent case)
	{
		Error = UlckErr;
	}
	return (Error);
}


//**** Check current input status ****
//Return true if a source is connected to the current input
//************************************
boolean Current_Input_Status(void)
{
	boolean Connected = false;
	byte Reg25 = AK4118_Read(0x25);

	switch (Current_Input)
	{
		case 0:
			if (Reg25 & B00000100)	//Status of AK4118 input2 : USB
			{
				Connected = true;
			}
			break;
		case 1:
			if (Reg25 & B00000001) 	//Status of AK4118 input0 : Optical
			{
				Connected = true;
			}
			break;
		case 2:
			if (Reg25 & B00000010) 	//Status of AK4118 input1 : Coax
			{
				Connected = true;
			}
			break;
		case 3:
			if (Reg25 & B00000010) 	//Status of AK4118 input1 : Coax
			{
				Connected = true;
			}
			break;
		case 4:
			if (Reg25 & B00000010) 	//Status of AK4118 input1 : Coax
			{
				Connected = true;
			}
			break;
		case 5:
			if (Reg25 & B00000010) 	//Status of AK4118 input1 : Coax
			{
				Connected = true;
			}
			break;
	}
	return (Connected);
}

//**** AK4118 Event detection function ****
//Reply with an Event Code corresponding to the event detected :
//Change of sampling frequency, AK4118 error or end of error condition.
//Event Code is No_Event when no error. Based on a four state engine.
//Event are detected by the ISR AK4118_Int()
//*******************************************
int AK4118_Event_Detect(void)
{
	byte Event = No_Event;
	byte Stream_Err;

	switch (Int_AK4118)
	{
		case Int_No:
			break;

		case Int_Trig:
			AK4118_Err_Code = Identify_AK4118_Error();
			if(AK4118_Err_Code == NoErr)
			{
				Int_AK4118 = Int_No;						//fast end process
			}
			else if (AK4118_Err_Code == SfErr)
			{
				AK4118_Err_Code = NoErr;
				Event = Ev_4118_SfChange;					//ask for sampling frequency update
				Int_AK4118 = Int_No;						//fast end process
			}
			else
			{
				Event = Ev_4118_Error;						//mute Dac
				Int_AK4118 = Int_Ongoing;
				Prev_AK4118_Error=AK4118_Err_Code;
			}
			break;

		case Int_Ongoing:
			delay(60);										//slow down main loop while in error ongoing status
			AK4118_Reg6_img = AK4118_Read(6);				//update status
			AK4118_Reg26_img = AK4118_Read(0x26);			//update status
			AK4118_Err_Code = Identify_AK4118_Error();		//update status
			if (AK4118_Err_Code == NoErr)					//stay in Ongoing status until all errors disappeared
			{
				Int_AK4118 = Int_End;
			}
			else if (AK4118_Err_Code == SfErr)
			{
				Event = Ev_4118_SfChange;					//ask for sampling frequency update
			}
			else if (AK4118_Err_Code != Prev_AK4118_Error)	//handle error change while in error status
			{
				refresh_MS=true;							//Display the updated error code
				Prev_AK4118_Error=AK4118_Err_Code;
			}
			break;

		case Int_End:
			if (!digitalRead(AK4118_INT0))					//wait INT0 output to return Low
			{
				Event = Ev_4118_End_Error;
				Int_AK4118 = Int_No;
			}
			break;
	}
	return (Event);
}

//**** End of Delay Event detection function ****
//Reply with an Event Code corresponding to delay expired
//Event Code is No_Event when no delay expiration
//*******************************************
int Delay_Event_Detect(void)
{
	int Event = No_Event;
	if (Delay3_Ctr != 0)							//check if Delay3 is running
	{
		if (Delay3_Ctr + Delay3_Val < millis())
		{
			Delay3_Ctr = 0;
			Event = Ev_De3;
		}
	}
	else if (Delay2_Ctr != 0)						//check if Delay2 is running
	{
		if (Delay2_Ctr + Delay2_Val < millis())
		{
			Delay2_Ctr = 0;
			Event = Ev_De2;
		}
	}
	else if (Delay1_Ctr != 0)						//check if Delay1 is running
	{
		if (Delay1_Ctr + Delay1_Val < millis())
		{
			Delay1_Ctr = 0;
			Event = Ev_De1;
		}
	}
	return(Event);
}

//**** AK4118 write ****
//**********************
void AK4118_Write(byte Reg, byte Value)
{
	digitalWrite(CSN, LOW);
	if(Reg > 0x1F)
	{
		Reg = Reg + 0x20;			//tweak to handle the weird adress bit mapping scheme
	}
	SPI.transfer(Reg | B00100000);	//set write bit to 1
	SPI.transfer(Value);
	digitalWrite(CSN, HIGH);
}

//**** AK4118 read ****
//**********************
byte AK4118_Read(byte Reg)
{
	byte Data;
	digitalWrite(CSN, LOW);
	if(Reg > 0x1F)
	{
		Reg = Reg + 0x20;			//tweak to handle the weird adress bit mapping scheme
	}
	SPI.transfer(Reg);				//write bit is is 0
	Data = SPI.transfer(None);
	digitalWrite(CSN, HIGH);
	return(Data);
}

//**** AK4495 write ****
//**********************
void AK4495_Write(byte Reg, byte Value)
{
	Wire.beginTransmission(AK4495_Addr);
	Wire.write(Reg);
	Wire.write(Value);
	Wire.endTransmission();
}


//**** AK4495 Mute ****
//**********************
void AK4495_Mute(void)
{
	AK4495_Reg1 |= B00000001;
	AK4495_Write(1, AK4495_Reg1);
}

//**** AK4495 UnMute ****
//**********************
void AK4495_UnMute(void)
{
	if (Int_AK4118 == Int_No & !Mute_SW)		//unmute if no error and Mute switch inactive
	{
		AK4495_Reg1 &= B11111110;
		AK4495_Write(1, AK4495_Reg1);
	}

}

//**** DAC Select Input ****
//Set the input (AK4118 + relay board) according to Current_Input global variable then
//set the corresponding volume
//**************************
void Set_Dac_In(void)
{
	static int Previous_Input = 0xFFFF;							//force update at power up only
	if (Current_Input != Previous_Input)
	{
		digitalWrite(Relay1, LOW);								//shut off spdif inputs and wait relays to move
		digitalWrite(Relay2, LOW);
		digitalWrite(Relay3, LOW);
		digitalWrite(Relay4, LOW);
		delay(10);
		switch (Current_Input)
		{
			case 0:
				AK4118_Write(3, AK4118_Reg3_I | B00000010 );	//IPS0,1,2 = 2 for input 2 (USB)
				break;
			case 1:
				AK4118_Write(3, AK4118_Reg3_I);					//IPS0,1,2 = 0 for input 0 (Opt)
				break;
			case 2:
				digitalWrite(Relay1, HIGH);
				AK4118_Write(3, AK4118_Reg3_I | B00000001);		//IPS0,1,2 = 1 for input 1 (Coax)
				break;
			case 3:
				digitalWrite(Relay2, HIGH);
				AK4118_Write(3, AK4118_Reg3_I | B00000001);		//IPS0,1,2 = 1 for input 1 (Coax)
				break;
			case 4:
				digitalWrite(Relay3, HIGH);
				AK4118_Write(3, AK4118_Reg3_I | B00000001);		//IPS0,1,2 = 1 for input 1 (Coax)
				break;
			case 5:
				digitalWrite(Relay4, HIGH);
				AK4118_Write(3, AK4118_Reg3_I | B00000001);		//IPS0,1,2 = 1 for input 1 (Coax)
				break;
		}
		Set_Dac_Vol();											//Set volume for the new input
		Current_TScreen = Current_Input; 						//trim volume screen to be prompted for current input
		Previous_Input = Current_Input;
	}
}

//**** Set DAC Output Volume ****
//Set the volume according to the global volume and the trim volume of the current Input
//*******************************
void Set_Dac_Vol(void)
{
	int Volume2set = DAC_Vol - Trim_Volume[Current_Input];
	if (Volume2set < -127 )
	{
		Volume2set = -127;
	}
	AK4495_Write(3, byte(2 * Volume2set + 255));
	AK4495_Write(4, byte(2 * Volume2set + 255));
}

//**** Set DAC Filter ****
//Set the filter to use according to the global variable Current_Filter
//*******************************
void Set_Dac_Filter(void)
{
	byte reg2, reg5;
	reg5 = AK4495_Reg5_I;						//SSLOW bit is 0
	switch(Current_Filter)
	{
		case 0:
			AK4495_Reg1 &= B11011111;			//set SD bit to 0
			reg2 = AK4495_Reg2_I;				//set SLOW bit to 0
			break;
		case 1:
			AK4495_Reg1 |= B00100000;			//set SD bit to 1
			reg2 = AK4495_Reg2_I;				//set SLOW bit to 0
			break;
		case 2:
			AK4495_Reg1 &= B11011111;			//set SD bit to 0
			reg2 = AK4495_Reg2_I | B00000001;	//set SLOW bit to 1
			break;
		case 3:
			AK4495_Reg1 |= B00100000;			//set SD bit to 1
			reg2 = AK4495_Reg2_I | B00000001;	//set SLOW bit to 1
			break;
		case 4:
			reg5 = AK4495_Reg5_I | B00000001;	//set SSLOW bit to 1
			break;
	}
	AK4495_Write(5, reg5);
	AK4495_Write(1, AK4495_Reg1);
	AK4495_Write(2, reg2);
}

//**** Set DAC Sound Quality ****
//Set the Sound Quality to use according to the global variable Current_SoundQ
//*******************************
void Set_Dac_SoundQ(void)
{
	AK4495_Write(8, Current_SoundQ);
}

//**** Main Screen ****
// Display the main status information
// Provide screen animation sequenced by Event EvDe3
// Store new settings in EEPROM when modified only
//*********************
void MainScreen(void)
{
	byte Temp_array[8];
	static byte Anim_Err [Anim_Lenght][2][8] =
	{
		{
			{	0x0,	0x0,	0x0,	0x0,	0x0,	0x0,	0x0,	0x0	},
			{	0x0,	0xE,	0x1F,	0x1F,	0x1B,	0x1F,	0x1F,	0xE	}
		},
		{
			{	0x0,	0x0,	0x0,	0x0,	0x4,	0x0,	0x0,	0x0	},
			{	0x0,	0x0,	0xE,	0x1F,	0x1B,	0x1F,	0xE,	0x0	}
		},
		{
			{	0x0,	0x0,	0x0,	0xE,	0xA,	0xE,	0x0,	0x0	},
			{	0x0,	0x0,	0x0,	0xE,	0xA,	0xE,	0x0,	0x0	}
		},
		{
			{	0x0,	0x0,	0xE,	0x1F,	0x1B,	0x1F,	0xE,	0x0	},
			{	0x0,	0x0,	0x0,	0x0,	0x4,	0x0,	0x0,	0x0	}
		},
		{
			{	0x0,	0xE,	0x1F,	0x1F,	0x1B,	0x1F,	0x1F,	0xE	},
			{	0x0,	0x0,	0x0,	0x0,	0x0,	0x0,	0x0,	0x0	}
		}
	};

	static byte Anim_Ok [Anim_Lenght + 1][4][8] =
	{
		{
			{	0x0,	0x0,	0x0,	0x0,	0x0,	0x0,	0x0,	0x1	},
			{	0x0,	0x0,	0x0,	0x0,	0x0,	0x0,	0x0,	0x10},
			{	0x1,	0x0,	0x0,	0x0,	0x0,	0x0,	0x0,	0x0	},
			{	0x10,	0x0,	0x0,	0x0,	0x0,	0x0,	0x0,	0x0	}
		},
		{
			{	0x0,	0x0,	0x0,	0x0,	0x0,	0x0,	0x3,	0x2	},
			{	0x0,	0x0,	0x0,	0x0,	0x0,	0x0,	0x18,	0x8	},
			{	0x2,	0x3,	0x0,	0x0,	0x0,	0x0,	0x0,	0x0	},
			{	0x8,	0x18,	0x0,	0x0,	0x0,	0x0,	0x0,	0x0	}
		},
		{
			{	0x0,	0x0,	0x0,	0x0,	0x0,	0x3,	0x4,	0x4	},
			{	0x0,	0x0,	0x0,	0x0,	0x0,	0x18,	0x4,	0x4	},
			{	0x4,	0x4,	0x3,	0x0,	0x0,	0x0,	0x0,	0x0	},
			{	0x4,	0x4,	0x18,	0x0,	0x0,	0x0,	0x0,	0x0	}
		},
		{
			{	0x0,	0x0,	0x0,	0x0,	0x7,	0x8,	0x8,	0x8	},
			{	0x0,	0x0,	0x0,	0x0,	0x1C,	0x2,	0x2,	0x2	},
			{	0x8,	0x8,	0x8,	0x7,	0x0,	0x0,	0x0,	0x0	},
			{	0x2,	0x2,	0x2,	0x1C,	0x0,	0x0,	0x0,	0x0	}
		},
		{
			{	0x0,	0x0,	0x0,	0xF,	0x10,	0x10,	0x10,	0x11},
			{	0x0,	0x0,	0x0,	0x1E,	0x1,	0x1,	0x1,	0x11},
			{	0x11,	0x10,	0x10,	0x10,	0xF,	0x0,	0x0,	0x0	},
			{	0x11,	0x1,	0x1,	0x1,	0x1E,	0x0,	0x0,	0x0	}
		},
		{
			//Those 4 last characters are not part of the 5 steps animations
			//but define the mute symbol to be displayed when Dac muted. It is
			//convenient to store them within this 3D array
			{	0x0,	0x0,	0x0,	0x0,	0x0,	0xF,	0xF,	0xF },
			{	0x1,	0x3,	0x7,	0xF,	0x1F,	0x1F,	0x1F,	0x1F},
			{	0xF,	0xF,	0xF,	0x0,	0x0,	0x0,	0x0,	0x0 },
			{	0x1F,	0x1F,	0x1F,	0x1F,	0xF,	0x7,	0x3,	0x1	}
		}
	};

	if (Previous_Screen != Current_Screen)
	{
		lcd.clear();
		Previous_Screen = Current_Screen;
		refresh_MS = true;									//ask for screen refresh
	}
	if(Save_NV_FS)
	{
		Save_NV_FS = false;
		if (Current_Filter != EEPROM.read(NV_Filter))
		{
			EEPROM.write(NV_Filter, Current_Filter);		//save filter setup in NV memory if changed only
		}
	}
	if(Save_NV_QS)
	{
		Save_NV_QS = false;
		if (Current_SoundQ != EEPROM.read(NV_SoundQ))
		{
			EEPROM.write(NV_SoundQ, Current_SoundQ);		//save sound quality setup in NV memory if changed only
		}
	}
	if(Save_NV_TS)
	{
		Save_NV_TS = false;
		for (int i = 0; i < 6; i++)
		{
			if(Trim_Volume[i] != EEPROM.read(NV_TVOL0 + i)) //store trim volume in NV if changed
			{
				EEPROM.write(NV_TVOL0 + i, Trim_Volume[i]);
			}
		}
	}
	if (refresh_MS)
	{
		lcd.setCursor(0, 0);
		lcd.print(Input_Name [Current_Input]);
		if (Pre_emph_status)
		{
			lcd.setCursor(13, 0);
			lcd.print("*");
		}
		lcd.setCursor(0, 1 );
		lcd.print("Filter ");
		lcd.print(Current_Filter);
		lcd.setCursor(9, 1);
		lcd.print("SQ ");
		lcd.print(Current_SoundQ + 1);
		lcd.setCursor(6, 0);
		if (Int_AK4118)										//true if an error is ongoing
		{
			lcd.print(Error_msg[AK4118_Err_Code]);
		}
		else
		{
			lcd.print(Sf_Display[Current_Freq]);
			lcd.setCursor(9, 0);
			lcd.print("kHz  ");
		}
		refresh_MS = false;
	}

//manage animation on MS screen
	if (Delay3_Ctr == 0)
	{
		Delay3_Ctr = millis();								//Re launch Delay 3 counter
		int i, j, k;
		for (i = 0; i < 4; ++i)
		{
			for (j = 0; j < 8; ++j)
			{
				if (Mute_SW)
				{
					k = 5;
				}
				else
				{
					k = Animation_ctr;					//when mute toggle switch on, Mute symbol is
					//displayed instead of animation
				}
				Temp_array[j] = Anim_Ok[k][i][j];		//copy the 8 bytes from the 3D array to the 1D one
			}
			lcd.createChar(i, Temp_array);				//create the 4 customized characters for each animation step (k)
		}
		for (int i = 0; i < 2; ++i)
		{
			for (int j = 0; j < 8; ++j)
			{
				Temp_array[j] = Anim_Err[Animation_ctr][i][j];	//copy the 8 bytes from the 3D array to the 1D one
			}
			lcd.createChar(i + 4, Temp_array);					//create the 2 customized characters for each animation step
		}
		if (Int_AK4118)											//If error is ongoing
		{
			lcd.setCursor(14, 1);								//if error status then display the error animation
			lcd.write(4);
			lcd.write(5);
		}
		else
		{
			lcd.setCursor(14, 0);								//otherwise display the running OK animation
			lcd.write(0);
			lcd.write(1);
			lcd.setCursor(14, 1);
			lcd.write(2);
			lcd.write(3);
		}
	}
}

//**** Volume Screen ****
//Change volume level
//***********************
void VolumeScreen(void)
{
	if (Previous_Screen != Current_Screen)
	{
		lcd.clear();
		Previous_Screen = Current_Screen;
		lcd.setCursor(1, 0);
		lcd.print("VOLUME ");
		lcd.setCursor(12, 0);
		lcd.print("dB");
		refresh_VS = true;
	}
	if (refresh_VS)
	{
		lcd.setCursor(8, 0);
		lcd.print("    ");
		lcd.setCursor(8, 0);
		lcd.print(DAC_Vol);		//Scale from -127dB to 0dB
		refresh_VS = false;
	}
}

//**** Filter Screen ****
//Change Filter used
//***********************
void FilterScreen(void)
{
	if (Previous_Screen != Current_Screen)
	{
		lcd.clear();
		Previous_Screen = Current_Screen;
		lcd.print("FILTER");
		refresh_FS = true;
	}
	if (refresh_FS)
	{
		lcd.setCursor(11, 0);
		lcd.print(Filter_Name[Current_Filter * 2]);
		lcd.setCursor(5, 1);
		lcd.print(Filter_Name[Current_Filter * 2 + 1]);
		refresh_FS = false;
	}
}

//**** Sound Quality Screen ****
//Change Sound Quality used
//******************************
void SoundQScreen(void)
{
	if (Previous_Screen != Current_Screen)
	{
		lcd.clear();
		Previous_Screen = Current_Screen;
		lcd.print("SOUND QUALITY");
		refresh_QS = true;
	}
	if (refresh_QS)
	{
		lcd.setCursor(13, 1);
		lcd.print("SQ");
		lcd.print(Current_SoundQ + 1);
		refresh_QS = false;
	}
}

//**** Trim Volume Screen ****
//Trim Volume per Input
//****************************

void TrimVScreen(byte Input)
{
	if (Previous_Screen != Current_Screen)
	{
		lcd.clear();
		Previous_Screen = Current_Screen;
		lcd.print("TRIM VOLUME");
		lcd.setCursor(0, 1);
		lcd.print("IN:");
		lcd.setCursor(13, 1);
		lcd.print("dB");
		refresh_TS = true;					//ask for screen refresh
	}
	if (refresh_TS)
	{
		lcd.setCursor(0, 4);
		lcd.print(Input_Name [Input]);
		lcd.setCursor(10, 1);
		lcd.print("   ");
		lcd.setCursor(10, 1);
		lcd.print(-1 * Trim_Volume[Input]);		//Scale from -12dB to 0dB
		refresh_TS = false;
	}
}

//**** Sf_Change function ****
//Update sampling freqency according AK4118 detection
//****************************
void Sf_Change(void)
{
	byte i = -1;
	boolean match = false;
	AK4118_Reg7_img = AK4118_Read(7) & B11110000;

	do
	{
		++i;
		//find the index (i) of the Sampling Frequency detected in the list of Frequency supported by the software
		//this index will allow to lookup the corresponding Master Clock generator setting and the text to Display on lcd.
		//index is stored in global variable: Current_Freq
		//
		match = AK4118_Reg7_img == AK4118_Sf[i];
	}
	while (!match && i < Nofreq);
	if (match)
	{
		Current_Freq = i;
		AK4118_Write(0, AK4118_MCK[i] | AK4118_Reg0_I);		//set MCK output for AK4495
		//Nothing to do on AK4495 side
	}
	else
	{
		Current_Freq = Nofreq;
	}
	Pre_emph_status = AK4118_Read(6) & B00000100;		//update pre emphasize status
	refresh_MS = true;
}

//**** Execute the Action corresponding to each Event in the context of each screen ****
//*
//****************************************************************************************
int Exec_Action ( int Event2Process)
{
	int * ScrEvPtr, * ScrAcPtr;	//define pointers targeting the arrays
	//which provides, per screen, the association between Events and Action
	int Action2Exec = Ac_Noac;
	static int Anim_Direction = 1;
	byte i = -1;
	boolean match = false;

	switch (Current_Screen)
	{
		case MS:
			ScrEvPtr = MS_Event;
			ScrAcPtr = MS_Action;
			break;
		case VS:
			ScrEvPtr = VS_Event;
			ScrAcPtr = VS_Action;
			break;
		case QS:
			ScrEvPtr = QS_Event;
			ScrAcPtr = QS_Action;
			break;
		case FS:
			ScrEvPtr = FS_Event;
			ScrAcPtr = FS_Action;
			break;
		case TS:
			ScrEvPtr = TS_Event;
			ScrAcPtr = TS_Action;
			break;
	}
	do
	{
		++i;
		//find the index (i) of the Event2process in the list of Events supported by the screen
		//this index will allow to lookup the corresponding Action
		match = *(ScrEvPtr + i) == Event2Process;
	}
	while (!match && (*(ScrEvPtr + i) != 0xFF));		//FF is end of list
	if (match)
	{
		Action2Exec = *(ScrAcPtr + i);					//Lookup the corresponding Action
	}
	switch (Action2Exec)
	{
		case Ac_SfChange:
			Sf_Change();
			break;

		case Ac_MS_anim:
			Animation_ctr = Animation_ctr + Anim_Direction;		//move animation one step
			if (Animation_ctr == 0)
			{
				Anim_Direction = 1;
			}
			else if(Animation_ctr == Anim_Lenght - 1)
			{
				Anim_Direction = -1;
			}
			break;

		case Ac_4118_Err:
			AK4495_Mute();
			refresh_MS = true;
			break;

		case Ac_4118_End_Err:
			AK4495_UnMute();			//free the sound !
			refresh_MS = true;
			break;

		case Ac_MS:
			Current_Screen = MS;		//Go to Main Screen
			break;

		case Ac_VS:
			Current_Screen = VS;		//Go to Volume Screen
			Delay1_Ctr = millis();		//launch delay1
			break;

		case Ac_QS:
			Current_Screen = QS;		//Go to Sound Quality Screen
			Delay1_Ctr = millis();		//launch delay1
			break;

		case Ac_FS:
			Current_Screen = FS;		//Go to Sound Filter Screen
			Delay1_Ctr = millis();		//launch delay1
			break;

		case Ac_TS:
			Current_Screen = TS;		//Go Volume Trim Screen
			Delay2_Ctr = millis();		//launch delay2
			break;

		case Ac_TSnext:
			++Current_TScreen;			//Move to next Trim Volume Screen
			if (Current_TScreen > 5)
			{
				Current_TScreen = 0;
			}
			Delay2_Ctr = millis();		//launch delay2
			refresh_TS = true;
			break;

		case Ac_VSplus:
			Current_Screen = VS;
			if (DAC_Vol > -21 && DAC_Vol < 0)
			{
				DAC_Vol += 1;
			}
			else if (DAC_Vol > -61 && DAC_Vol < -20)
			{
				DAC_Vol += 2;
			}
			else if (DAC_Vol > -129 && DAC_Vol < -60)
			{
				DAC_Vol += 4;
			}
			Set_Dac_Vol();
			Delay1_Ctr = millis();		//launch delay1
			refresh_VS = true;
			break;

		case Ac_VSminus:
			Current_Screen = VS;
			if (DAC_Vol > -20 && DAC_Vol <= 0)
			{
				DAC_Vol -= 1;
			}
			else if (DAC_Vol > -60 && DAC_Vol < -19)
			{
				DAC_Vol -= 2;
			}
			else if (DAC_Vol > -128 && DAC_Vol < -59)
			{
				DAC_Vol -= 4;
			}
			Set_Dac_Vol();
			Delay1_Ctr = millis();		//launch delay1
			refresh_VS = true;
			break;

		case Ac_QSplus:
			++Current_SoundQ;
			if (Current_SoundQ > 4)
			{
				Current_SoundQ = 0;
			}
			Set_Dac_SoundQ();
			Save_NV_QS = true;
			Delay1_Ctr = millis();		//launch delay1
			refresh_QS = true;
			break;

		case Ac_QSminus:
			--Current_SoundQ;
			if (Current_SoundQ < 0)
			{
				Current_SoundQ = 4;
			}
			Set_Dac_SoundQ();
			Save_NV_QS = true;
			Delay1_Ctr = millis();		//launch delay1
			refresh_QS = true;
			break;

		case Ac_FSplus:
			++Current_Filter;
			if (Current_Filter > 4)
			{
				Current_Filter = 0;
			}
			Set_Dac_Filter();
			Save_NV_FS = true;
			Delay1_Ctr = millis();		//launch delay1
			refresh_FS = true;
			break;

		case Ac_FSminus:
			--Current_Filter;
			if (Current_Filter < 0)
			{
				Current_Filter = 4;
			}
			Set_Dac_Filter();
			Save_NV_FS = true;
			Delay1_Ctr = millis();		//launch delay1
			refresh_FS = true;
			break;

		case Ac_SSplus:
			++Current_Input;
			if (Current_Input > 5)
			{
				Current_Input = 0;
			}
			Set_Dac_In();				//Set input
			refresh_MS = true;
			break;

		case Ac_SSminus:
			--Current_Input;
			if (Current_Input < 0)
			{
				Current_Input = 5;
			}
			Set_Dac_In();				//Set input
			refresh_MS = true;
			break;

		case Ac_In0:
			Current_Input = 0;
			Set_Dac_In(); 				//Set input
			refresh_MS = true;
			break;

		case Ac_In1:
			Current_Input = 1;
			Set_Dac_In();				//Set input
			refresh_MS = true;
			break;

		case Ac_In2:
			Current_Input = 2;
			Set_Dac_In();				//Set input
			refresh_MS = true;
			break;

		case Ac_In3:
			Current_Input = 3;
			Set_Dac_In();				//Set input
			refresh_MS = true;
			break;

		case Ac_In4:
			Current_Input = 4;
			Set_Dac_In();				//Set input
			refresh_MS = true;
			break;

		case Ac_In5:
			Current_Input = 5;
			Set_Dac_In();				//Set input
			refresh_MS = true;
			break;

		case Ac_TSplus:
			if (Trim_Volume[Current_TScreen] < Max_Trim)
			{
				++Trim_Volume[Current_TScreen];
			}
			Set_Dac_Vol();						//Set volume
			Delay2_Ctr = millis();				//launch delay2
			Save_NV_TS = true;
			refresh_TS = true;
			break;

		case Ac_TSminus:
			if (Trim_Volume[Current_TScreen] > 0)
			{
				--Trim_Volume[Current_TScreen];
			}
			Set_Dac_Vol();						//Set volume
			Delay2_Ctr = millis();				//launch delay2
			Save_NV_TS = true;
			refresh_TS = true;
			break;

		case Ac_Mute_Toggle:
			if (!Int_AK4118)					//toggle only when no ongoing error
			{
				Mute_SW = !Mute_SW;				//toggle mute status
				if (Mute_SW)
				{
					AK4495_Mute();
				}
				else
				{
					AK4495_UnMute();
				}
				refresh_MS = true;
			}
			break;

		case Ac_Debug:
			break;
	}
}


//**** Main program ****
//
//
//**********************

void loop()
{
	dummy = Delay_Event_Detect();
	if (dummy)
	{
		Exec_Action(dummy);
	}
	dummy = IR_Event_Detect();
	if (dummy)
	{
		Exec_Action (dummy);
	}
	dummy = Button_Event_Detect();
	if (dummy)
	{
		Exec_Action(dummy);
	}
	dummy = AK4118_Event_Detect();
	if (dummy)
	{
		Exec_Action(dummy);
	}
	switch (Current_Screen)
	{
		case MS:
			MainScreen();
			break;
		case VS:
			VolumeScreen();
			break;
		case FS:
			FilterScreen();
			break;
		case QS:
			SoundQScreen();
			break;
		case TS:
			TrimVScreen(Current_TScreen);
			break;
	}
}


//***************************************************************************************
//***************************************************************************************
//******************************** END OF CODE ******************************************
